""" Read bam files"""
import copy

import mirtop.libs.logger as mylog

logger = mylog.getLogger(__name__)


def _coord(sequence, start, mirna, precursor, iso):
    """
    Define t5 and t3 isomirs
    """
    insertion = 0
    deletion = 0
    add = 0
    if iso.subs:
        insertion = sum([1  if s[-1] == "-" else 0 for s in iso.subs])
    if iso.subs:
        deletion = sum([1  if s[1] == "-" else 0 for s in iso.subs])
    if iso.add:
        add = len(iso.add)
    end = (iso.end - add - insertion + deletion)

    logger.debug("COOR:: s:%s len:%s end:%s fixedEnd:%s mirna:%s iso:%s" % (
        start, len(sequence), iso.end, end, mirna, iso.format())
        )
    dif = abs(mirna[0] - start)
    if start < mirna[0]:
        iso.t5 = sequence[:dif].upper()
    elif start > mirna[0]:
        iso.t5 = precursor[mirna[0]:mirna[0] + dif].lower()
    elif start == mirna[0]:
        iso.t5 = 0
    if dif > 6:
        logger.debug("COOR::start > 6 start:%s len:%s dif:%s mirna:%s iso:%s" % (
                        start, len(sequence),
                        dif, mirna, iso.format()))
        return None

    dif = abs(mirna[1] - iso.end)
    if iso.add:
        iso.add = iso.add.replace("-", "")
        sequence = sequence[:-len(iso.add)]
    if end > mirna[1]:
        iso.t3 = sequence[-dif:].upper()
    elif end < mirna[1]:
        iso.t3 = precursor[mirna[1] + 1 - dif:(mirna[1] + 1)].lower()
    elif end == mirna[1]:
        iso.t3 = 0

    if dif > 7:
        logger.debug("COOR::end > 7 len:%s end:%s dif:%s mirna:%s iso:%s" % (
            len(sequence), end, dif, mirna, iso.format()))
        return None
    return True


def annotate(reads, mature_ref, precursors, quiet=False):
    """
    Using coordinates, mismatches and realign to annotate isomiRs

    Args:
        *reads(dicts of hits)*:
            dict object that comes from *mirotp.bam.bam.read_bam()*

        *mirbase_ref (dict of mirna positions)*:
            dict object that comers from *mirtop.mirna.read_mature()*

        *precursors dict object (key : fasta)*:
            that comes from *mirtop.mirna.fasta.read_precursor()*
        *quiet(boolean)*:
            verbosity state
    Return:
        *reads (dict)*:
            dictionary where keys are read_id and
            values are *mirtop.realign.hits*
    """
    n_iso = 0
    n_skip_precursor = 0
    for r in reads:
        logger.debug(("\nANN::READ::read {r}").format(**locals()))
        for ps in reads[r].precursors:
            p = list(ps)[0]
            start = reads[r].precursors[ps].start
            end = reads[r].precursors[ps].end
            logger.debug(("\nANN::READ::precursor {start} {end}").format(**locals()))
            for mature in mature_ref[p]:
                mi = mature_ref[p][mature]
                logger.debug(("\nANN::NEW::read:{s}\n pre:{p} start:{start} end: {end} "
                              "cigar: {cigar} "
                              "\n mir:{mature} mir_pos:{mi}\n mir_seqs:{mature_s}"
                              ).format(s=reads[r].sequence,
                                       mature_s = precursors[ps][mi[0]:mi[1] + 1],
                                       cigar = reads[r].precursors[ps].cigar,
                                       **locals()))
                iso_copy = copy.deepcopy(reads[r].precursors[ps])
                if not precursors[ps[0]]:
                    n_skip_precursor += 1
                    continue
                is_iso = _coord(reads[r].sequence, start, mi, precursors[ps[0]], iso_copy)
                logger.debug(("ANN::is_iso:{is_iso}").format(**locals()))
                logger.debug("ANN::annotation:%s iso:%s" % (r, reads[r].precursors[ps].format()))
                logger.debug("ANN::annotation:%s Variant:%s" % (r, reads[r].precursors[ps].formatGFF()))
                if is_iso:
                    n_iso += 1
                    reads[r].precursors[ps] = iso_copy
                    reads[r].precursors[ps].mirna = mature
                    # break
    if not quiet:
        logger.info("Valid hits (+/- reference miRNA): %s" % n_iso)
        logger.info("Skipped due to not precursor sequence: %s" % n_skip_precursor)
    return reads



"""Helpers to work with fastq files"""

import os
from itertools import product
import gzip


def open_fastq(in_file):
    """ open a fastq file, using gzip if it is gzipped
        (from bcbio package)

    Args:
        *in_file(str)*: file name.

    Returns:
        *(File)*: file handler.
    """
    _, ext = os.path.splitext(in_file)
    if ext == ".gz":
        return gzip.open(in_file, 'rb')
    if ext in [".fastq", ".fq", ".fasta", ".fa"]:
        return open(in_file, 'r')
    return ValueError("File needs to be fastq|fasta|fq|fa [.gz]")


def is_fastq(in_file):
    """Check whether file is fastq accepting
        txt, fq and fastq extensions understanding
        compression with gzip: .gzip and .gz
        (copy from bcbio)
    Args:
        *in_file(str)*: file name.

    Returns:
        *(boolean)*: Yes or Not.
    """
    fastq_ends = [".txt", ".fq", ".fastq"]
    zip_ends = [".gzip", ".gz"]
    base, first_ext = os.path.splitext(in_file)
    second_ext = os.path.splitext(base)[1]
    if first_ext in fastq_ends:
        return True
    elif (second_ext, first_ext) in product(fastq_ends, zip_ends):
        return True
    else:
        return False


def splitext_plus(fn):
    """Split on file extensions, allowing for zipped extensions.
        (copy from bcbio)

    Args:
        *fn(str)*: file name.

    Returns:
        *base, ext(str, str)*: basename and extesion.
    """
    base, ext = os.path.splitext(fn)
    if ext in [".gz", ".bz2", ".zip"]:
        base, ext2 = os.path.splitext(base)
        ext = ext2 + ext
    return base, ext



""" Read isomiR GFF files from optimir tool"""

from collections import defaultdict

import mirtop.libs.logger as mylog
from mirtop.gff.header import read_samples
from mirtop.gff.body import paste_columns
from mirtop.gff.body import variant_with_nt
from mirtop.gff.classgff import feature

logger = mylog.getLogger(__name__)


def read_file(fn, args):
    """
    Read OptimiR file and convert to mirtop GFF format.

    Args:
        *fn(str)*: file name with isomiR-SEA output information.

        *database(str)*: database name.

        *args(namedtuple)*: arguments from command line.
            See *mirtop.libs.parse.add_subparser_gff()*.

    Returns:
        *reads (nested dicts)*:gff_list has the format as
            defined in *mirtop.gff.body.read()*.

    """
    database = args.database
    gtf = args.gtf
    sep = " " if args.out_format == "gtf" else "="
    sample = read_samples(fn)
    reads = defaultdict(dict)
    logger.debug("OPTIMIR::SAMPLE::%s" % sample)
    with open(fn) as handle:
        for line in handle:
            gff = feature(line)
            fixed_line = line
            if gff.columns:
                if "Variant" not in gff.attributes:
                    gff.attributes["Variant"] = "NA"

                logger.debug("OPTIMIR::Chrom update from %s to %s" % (gff.columns["chrom"], gff.attributes["Parent"]))
                gff.columns["chrom"] = gff.attributes["Parent"].split(",")[0]
                fixed_line = gff.paste_columns(sep=sep)
                if args.add_extra:
                    extra = variant_with_nt(fixed_line, args.precursors, args.matures)
                    fixed_line = "%s Changes %s;" % (fixed_line, extra)

                fixed_line = paste_columns(feature(fixed_line), sep=sep)
                counts = gff.attributes["Expression"].split(",")
                chrom = gff.columns["chrom"]
                start = gff.columns["start"]
                if start not in reads[chrom]:
                    reads[chrom][start] = []
                reads[chrom][start].append([gff.attributes["UID"],
                                            gff.columns["chrom"],
                                            counts,
                                            sample, fixed_line])
    return reads



""" Read bam files"""
from __future__ import print_function
# from memory_profiler import profile

import os.path as op
import os
import pysam
from collections import defaultdict

import pybedtools

from mirtop.libs import do
from mirtop.libs.utils import file_exists
import mirtop.libs.logger as mylog
from mirtop.mirna.realign import isomir, hits, reverse_complement
from mirtop.mirna.mapper import get_primary_transcript, guess_database
from mirtop.bam import filter
from mirtop.gff import body
from mirtop.mirna.annotate import annotate
from mirtop.libs import sql

logger = mylog.getLogger(__name__)

# fp = open('memory_profiler.log', 'w+')


# @profile(stream=fp)
def read_bam(bam_fn, args, clean=True):
    """
    Read bam file and perform realignment of hits

    Args:
        *bam_fn*: a BAM file with alignments to the precursor

        *precursors*: dict with keys being precursor names and values
            being sequences. Come from mirtop.mirna.fasta.read_precursor().

        *clean*: Use mirtop.filter.clean_hits() to remove lower score hits.

    Returns:
        *reads (dict)*:
             keys are read_id and values are *mirtop.realign.hits*

    """
    bam_fn = _sam_to_bam(bam_fn)
    bam_fn = _bam_sort(bam_fn)
    reads = defaultdict(hits)
    if args.genomic:
        logger.warning("This is under development and variants can be unexact.")
        bed_fn = os.path.join(args.out, os.path.basename(bam_fn) + ".bed")
        logger.info("Making bed file.")
        _bed(bam_fn, bed_fn)
        logger.info("Intersecting bed file.")
        intersect_fn = intersect(bed_fn, args.gtf)
        # logger.info("Analyzing hits.")
        # reads = _read_lifted_bam(intersect_fn, reads, args, clean)
        logger.info("Loading database.")
        conn = _read_lifted_bam_alpha(intersect_fn, bam_fn, args)
        rows = sql.select_all_reads(conn)
        logger.info("Analyzing database.")
        precursors = args.precursors
        database = guess_database(args)
        reads = _read_lifted_lines(rows, precursors, database)
        conn.close()
    else:
        reads = _read_original_bam(bam_fn, reads, args, clean)
    logger.info("Done.")
    return reads


# @profile(stream=fp)
def low_memory_bam(bam_fn, sample, out_handle, args):
    if args.genomic:
        raise ValueError("low-memory option is not compatible with genomic coordinates.")
    precursors = args.precursors
    bam_fn = _sam_to_bam(bam_fn)
    bam_fn = _bam_sort(bam_fn)
    mode = "r" if bam_fn.endswith("sam") else "rb"
    handle = pysam.Samfile(bam_fn, mode)
    lines = []
    current = None
    for line in handle:
        if not current or current == line.query_name:
            lines.append(line)
            current = line.query_name
        else:
            reads = _read_lines(lines, precursors, handle, args)
            ann = annotate(reads, args.matures, args.precursors, quiet=True)
            gff_lines = body.create(ann, args.database, sample, args, quiet=True)
            body.write_body_on_handle(gff_lines, out_handle)
            current = line.query_name
            lines = []
            lines.append(line)
    reads = _read_lines(lines, precursors, handle, args)
    ann = annotate(reads, args.matures, args.precursors, quiet=True)
    gff_lines = body.create(ann, args.database, sample, args, quiet=True)
    body.write_body_on_handle(gff_lines, out_handle)


def low_memory_genomic_bam(bam_fn, sample, out_handle, args):
    logger.info("Reading BAM file in low memory mode.")
    logger.warning("This is under development and variants can be unexact.")
    precursors = args.precursors
    bam_fn = _sam_to_bam(bam_fn)
    bam_fn = _bam_sort(bam_fn)
    database = guess_database(args)
    bed_fn = os.path.join(args.out, os.path.basename(bam_fn) + ".bed")
    logger.info("Making bed file.")
    _bed(bam_fn, bed_fn)
    logger.info("Intersecting bed file.")
    intersect_fn = intersect(bed_fn, args.gtf)
    logger.info("Loading database.")
    conn = _read_lifted_bam_alpha(intersect_fn, bam_fn, args)
    rows = sql.select_all_reads(conn)
    lines = []
    current = None
    logger.info("Analyzing database.")
    for row in rows:
        if not current or current == row[0]:
            lines.append(row)
            current = row[0]
        else:
            reads = _read_lifted_lines(lines, precursors, database)
            ann = annotate(reads, args.matures, args.precursors, quiet=True)
            gff_lines = body.create(ann, args.database, sample, args, quiet=True)
            body.write_body_on_handle(gff_lines, out_handle)
            current = row[0]
            lines = []
            lines.append(row)
    reads = _read_lifted_lines(lines, precursors, database)
    ann = annotate(reads, args.matures, args.precursors, quiet=True)
    gff_lines = body.create(ann, args.database, sample, args, quiet=True)
    body.write_body_on_handle(gff_lines, out_handle)
    conn.close()
    logger.info("Done")


def _analyze_line(line, reads, precursors, handle, args):
    if line.reference_id < 0:
        logger.debug("READ::Sequence not mapped: %s" % line.reference_id)
        return reads
    if not line.cigarstring:
        logger.debug("READ::Sequence malformed: %s" % line)
        return reads
    query_name = line.query_name
    if query_name not in reads and not line.query_sequence:
        return reads
    sequence = line.query_sequence if not line.is_reverse else reverse_complement(line.query_sequence)
    logger.debug(("READ::Read name:{0} and Read sequence:{1}").format(line.query_name, sequence))
    if line.query_sequence and line.query_sequence.find("N") > -1:
        return reads
    if query_name not in reads:
        reads[query_name].set_sequence(sequence)
        reads[query_name].counts = _get_freq(query_name)
    if line.is_reverse and not args.genomic:
        logger.debug("READ::Sequence is reverse: %s" % line.query_name)
        return reads
    chrom = handle.getrname(line.reference_id)
    start = line.reference_start
    # If genomic endcode, liftover to precursor position
 #   if not start:
 #       logger.debug(("READ::not start found %s" % line.reference_start))
 #       return reads
    cigar = line.cigartuples
    # if line.cigarstring.find("I") > -1:
    #     indels_skip += 1
    iso = isomir()
    iso.align = line
    iso.set_pos(start, len(reads[query_name].sequence))
    logger.debug("READ::From BAM start %s end %s at chrom %s" % (iso.start, iso.end, chrom))
    if len(precursors[chrom]) < start + len(reads[query_name].sequence):
        logger.debug("READ::%s start + %s sequence size are bigger than"
                     " size precursor %s" % (
                                             line.reference_id,
                                             len(reads[query_name].sequence),
                                             len(precursors[chrom])))
        return reads
    iso.subs, iso.add, iso.cigar = filter.tune(
        reads[query_name].sequence, precursors[chrom],
        start, cigar)
    logger.debug("READ::After tune start %s end %s" % (iso.start, iso.end))
    logger.debug("READ::iso add %s iso subs %s" % (iso.add, iso.subs))
    reads[query_name].set_precursor(chrom, iso)
    return reads


def _read_lines(lines, precursors, handle, args, clean=True):
    reads = defaultdict(hits)
    for line in lines:
        reads = _analyze_line(line, reads, precursors, handle, args)
    if clean:
        reads = filter.clean_hits(reads)
    return reads


def _read_lifted_bam_alpha(bed_fn, bam_fn, args):
    database = guess_database(args)
    conn = sql.create_connection()
    key = "name" if args.keep_name else "sequence"
    sql.create_reads_table(conn, key)
    cur = conn.cursor()
    counts = 0
    seen = set()
    for line in bed_fn:
        fields = _parse_intersect(line, database, bed=True)
        if fields:
            hit = ".".join(fields[:3])
            if hit not in seen:
                counts += 1
                sql.insert_row_in_reads_table(cur, fields)
                seen.add(hit)
        # if counts == 1000:
        #     counts = 0
    del(hit)
    logger.info("Read %s lines that intersected with miRNAs." % counts)
    conn.commit()
    return conn


def _read_original_bam(bam_fn, reads, args, clean):
    mode = "r" if bam_fn.endswith("sam") else "rb"
    handle = pysam.Samfile(bam_fn, mode)
    indels_skip = 0
    precursors = args.precursors
    for line in handle:
        reads = _analyze_line(line, reads, precursors, handle, args)
    logger.info("Hits: %s" % len(reads))
    logger.info("Hits with indels %s" % indels_skip)
    if clean:
        reads = filter.clean_hits(reads)
        logger.info("Hits after clean: %s" % len(reads))
    return reads


def _parse_intersect(line, database, bed=False):
    "Parse bedtools intersect between bam_bed file and gtf from database"
    start_idx = 9 if bed else 15
    end_idx = 10 if bed else 16
    attr_idx = 14 if bed else 20
    if str(line).find(get_primary_transcript(database)) < 0: # only working with mirbase
        return None
    query_name = line[3]
    sequence = line[4]
    chrom = line[attr_idx].strip().split("Name=")[-1]
    start = line[1]
    end = line[2]
    strand = line[5]
    if not start:
        return None
    if strand == "+":
        start = int(start) - int(line[start_idx]) + 1
    else:
        start = int(line[end_idx]) - int(end)
    return (query_name, sequence, chrom, start)


def _analyze_lifted_line(line, reads, precursors, database):
    query_name = line[0]
    sequence = line[1]
    logger.debug(("READ::line name:{0}").format(line))
    if sequence and sequence.find("N") > -1:
        return reads
    if query_name not in reads:
        reads[query_name].set_sequence(sequence)
        reads[query_name].counts = _get_freq(query_name)
        reads[query_name].sequence = sequence

    chrom = line[2]
    start = line[3]
    iso = isomir()
    iso.align = line
    iso.set_pos(start, len(reads[query_name].sequence))
    logger.debug("READ::From BAM start %s end %s at chrom %s" % (iso.start, iso.end, chrom))
    if len(precursors[chrom]) < start + len(reads[query_name].sequence):
        logger.debug("READ::%s start + %s sequence size are bigger than"
                     " size precursor %s" % (
                                             chrom,
                                             len(reads[query_name].sequence),
                                             len(precursors[chrom])))
        return reads
    iso.subs, iso.add, iso.cigar = filter.tune(
        reads[query_name].sequence, precursors[chrom],
        start, None)
    logger.debug("READ::After tune start %s end %s" % (iso.start, iso.end))
    logger.debug("READ::iso add %s iso subs %s" % (iso.add, iso.subs))

    reads[query_name].set_precursor(chrom, iso)
    return reads


def _read_lifted_lines(lines, precursors, database, clean=True):
    reads = defaultdict(hits)
    for line in lines:
        reads = _analyze_lifted_line(line, reads, precursors, database)
    if clean:
        reads = filter.clean_hits(reads)
    return reads


def _read_lifted_bam(handle, reads, args, clean):
    indels_skip = 0
    precursors = args.precursors
    database = guess_database(args)
    for line in handle:
        rows = _parse_intersect(line, database, bed=True)
        reads = _analyze_lifted_line(rows, reads, precursors, database)
    logger.info("Hits: %s" % len(reads))
    logger.info("Hits with indels %s" % indels_skip)
    if clean:
        reads = filter.clean_hits(reads)
        logger.info("Hits after clean: %s" % len(reads))
    return reads


def _sam_to_bam(bam_fn):
    if not bam_fn.endswith("bam"):
        bam_out = "%s.bam" % os.path.splitext(bam_fn)[0]
        cmd = "samtools view -Sbh {bam_fn} -o {bam_out}"
        do.run(cmd.format(**locals()))
        return bam_out
    return bam_fn


def _bam_sort(bam_fn):
    bam_sort_by_n = op.splitext(bam_fn)[0] + "_sort.bam"
    if not file_exists(bam_sort_by_n):
        do.run(("samtools sort -n -o {bam_sort_by_n} {bam_fn}").format(
            **locals()))
    return bam_sort_by_n


def _get_freq(name):
    """
    Check if name read contains counts (_xNumber)
    """
    try:
        counts = int(name.split("_x")[1])
    except:
        return 0
    return counts


def _bed(bam_fn, bed_fn):
    mode = "r" if bam_fn.endswith("sam") else "rb"
    handle = pysam.Samfile(bam_fn, mode)
    current = None
    if os.path.exists(bed_fn):
        return bed_fn
    with open(bed_fn, 'w') as outh:
        for line in handle:
            if line.reference_id < 0:
                logger.debug("READ::Sequence not mapped: %s" % line.reference_id)
                continue
            if not line.cigarstring:
                logger.debug("READ::Sequence malformed: %s" % line)
                continue
            query_name = line.query_name
            if (not current or query_name!=current) and not line.query_sequence:
                continue
            if not current or query_name!=current:
                sequence = line.query_sequence if not line.is_reverse else reverse_complement(line.query_sequence)
            logger.debug(("READ::Read name:{0} and Read sequence:{1}").format(line.query_name, sequence))
            if line.query_sequence and line.query_sequence.find("N") > -1:
                continue
            chrom = handle.getrname(line.reference_id)
            start = line.reference_start
            end = start + len(sequence) - 1
            current = query_name
            strand = "+" if not line.is_reverse else "-"
            bed_line = "\t".join(list(map(str, [chrom, start, end, query_name, sequence, strand])))
            outh.write(bed_line + '\n')


def intersect(bam, gtf):
    bampy = pybedtools.BedTool(bam)
    gtfpy = pybedtools.BedTool(gtf)
    return bampy.intersect(gtfpy, wo=True, bed=True, s=True)


# deprecated from now on
def _read_quick_lines(lines):
    reads = defaultdict(hits)
    for line in lines:
        _analyze_quick_line(line, reads)
    return reads


def _analyze_quick_line(line, reads):
    if line.reference_id < 0:
        logger.debug("READ::Sequence not mapped: %s" % line.reference_id)
        return reads
    if not line.cigarstring:
        logger.debug("READ::Sequence malformed: %s" % line)
        return reads
    query_name = line.query_name
    if query_name not in reads and not line.query_sequence:
        return reads
    sequence = line.query_sequence if not line.is_reverse else reverse_complement(line.query_sequence)
    logger.debug(("READ::Read name:{0} and Read sequence:{1}").format(line.query_name, sequence))
    if line.query_sequence and line.query_sequence.find("N") > -1:
        return reads
    if query_name not in reads:
        reads[query_name].set_sequence(sequence)
        reads[query_name].counts = _get_freq(query_name)
    return reads


def _read_quick_bam(bam_fn, reads):
    mode = "r" if bam_fn.endswith("sam") else "rb"
    handle = pysam.Samfile(bam_fn, mode)
    for line in handle:
        reads = _analyze_quick_line(line, reads)
    return reads



""" Read GFF files and output FASTA format"""
from __future__ import print_function

import os
import sys

import mirtop.libs.logger as mylog
from mirtop.gff.classgff import feature
from mirtop.mirna.realign import read_id

logger = mylog.getLogger(__name__)


def convert(args):
    """
    Main function to convert from GFF3 to FASTA format.

    Args:
      *args*: supported options for this sub-command.
        See *mirtop.libs.parse.add_subparser_export()*.
    """
    for fn in args.files:
        logger.info("Reading %s" % fn)
        _process(fn, args.out)


def _process(fn, out_dir):
    if out_dir:
        out_fasta = os.path.join(out_dir, "%s.fasta" %
                                 os.path.splitext(os.path.basename(fn))[0])
    outh = sys.stdout if not out_dir else open(out_fasta, 'w')
    with open(fn) as inh:
        for line in inh:
            if line.startswith("#"):
                continue
            gff = feature(line)
            attr = gff.attributes
            read = read_id(attr["UID"])
            print((">{0}\n{1}").format(attr["UID"], read), file=outh)



""" Read prost! files"""

import os
from collections import defaultdict

from mirtop.mirna import mapper
import mirtop.libs.logger as mylog
from mirtop.mirna.realign import isomir, hits, get_mature_sequence, align
from mirtop.bam import filter

logger = mylog.getLogger(__name__)


def header():
    """
    Custom header for PROST! importer.

    Returns:
        *(str)*: PROST! header string.
    """

    return ""


def read_file(fn, hairpins, database, mirna_gtf):
    """
    Read PROST! file and convert to mirtop GFF format.

    Args:
        *fn(str)*: file name with PROST output information.

        *database(str)*: database name.

        *args(namedtuple)*: arguments from command line.
            See *mirtop.libs.parse.add_subparser_gff()*.

    Returns:
        *reads*: dictionary where keys are read_id and values are *mirtop.realign.hits*

    """
    reads = defaultdict(hits)
    sample = os.path.splitext(os.path.basename(fn))[0]
    genomics = mapper.read_gtf_to_mirna(mirna_gtf)
    matures = mapper.read_gtf_to_precursor(mirna_gtf)
    non_mirna = 0
    non_chromosome_mirna = 0
    outside_mirna = 0
    lines_read = 0
    ann, ann_type = _group_seqs_by_ann(fn)
    with open(fn) as handle:
        handle.readline()
        for line in handle:
            lines_read += 1
            cols = line.strip().split("\t")
            query_name = cols[0]
            query_sequence = cols[0]
            if not ann[query_sequence]:
                non_mirna += 1
                continue
            miRNA = ann_type[ann[query_sequence]][1]
            preNames = ann_type[ann[query_sequence]][0]
            if query_name not in reads and not query_sequence:
                continue
            if query_sequence and query_sequence.find("N") > -1:
                continue
            reads[query_name].set_sequence(query_sequence)
            reads[query_name].counts = cols[9]
            for preName in preNames.split(","):
                if preName in reads[query_name].precursors:
                    continue
                if preName not in hairpins:
                    non_chromosome_mirna += 1
                    continue
                reference_start = _align_to_mature(query_sequence, hairpins[preName], matures[preName][miRNA])
                logger.debug("\nPROST!::NEW::query: {query_sequence}\n"
                             "  precursor {preName}\n"
                             "  name:  {query_name}\n"
                             "  reference_start: {reference_start}\n"
                             "  mirna: {miRNA}".format(**locals()))
                iso = isomir()
                iso.align = line
                iso.set_pos(reference_start, len(reads[query_name].sequence))
                logger.debug("PROST!:: start %s end %s" % (iso.start, iso.end))
                if len(hairpins[preName]) < reference_start + len(reads[query_name].sequence):
                    continue
                iso.subs, iso.add, iso.cigar = filter.tune(
                    reads[query_name].sequence,
                    hairpins[preName],
                    reference_start, None)
                logger.debug("PROST!::After tune start %s end %s" % (
                    iso.start, iso.end))
                if len(iso.subs) < 2:
                    reads[query_name].set_precursor(preName, iso)
    logger.info("Lines loaded: %s" % lines_read)
    logger.info("Skipped lines because non miRNA in line: %s" % non_mirna)
    logger.info("Skipped lines because non chromosome in GTF:"
                " %s" % non_chromosome_mirna)
    logger.info("Skipped lines because outside precursor: %s" % outside_mirna)
    logger.info("Hits: %s" % len(reads))
    return reads


def _group_seqs_by_ann(fn):
    """Read file once to group sequences to same miRNA sequence"""
    ann = dict()
    ann_type = defaultdict(list)
    with open(fn) as inh:
        inh.readline()
        for line in inh:
            cols = line.strip().split("\t")
            ann[cols[0]] = cols[4]
            mirna = cols[11] if cols[11] else cols[13]
            hairpin = cols[15]
            if not cols[4] in ann_type:
                ann_type[cols[4]] = ["", ""]
            if mirna:
                ann_type[cols[4]][1] = mirna
            if hairpin:
                ann_type[cols[4]][0] = hairpin
    return [ann, ann_type]


def _align_to_mature(seq, hairpin, mature):
    """Get alignment between seq and mature"""
    mirna = get_mature_sequence(hairpin, mature)
    hit = align(seq, mirna)
    start = hit[0][:8].count("-") - 4 + int(mature[0])
    logger.debug("PROST::align:sequence to mature %s" % hit[0])
    logger.debug("PROST::align:start: %s -> %s" % (mature[0], start))
    return start


def _cigar_to_variants(seq, mature, cigar):
    """From mature based cigar get variants"""
    return None


def _make_variant(cols):
    logger.debug("PROST::variant: %s" % cols)
    variant = []
    if cols[0] != "0":
        variant.append("iso_5p:%s" % -1 * int(cols[0]))
    if cols[1] != "0":
        variant.append("iso_3p:%s" % cols[1])
    if cols[2] != "0":
        variant.append("iso_add3p:%s" % int(cols[2]))
    if cols[3] == "True":
        variant.append("iso_snv_seed")
    if cols[4] == "True":
        variant.append("iso_snv_central_offset")
    if cols[5] == "True":
        variant.append("iso_snv_central")
    if cols[6] == "True":
        variant.append("iso_snv_supp")
    if cols[7] == "True":
        variant.append("iso_snv")
    if not variant:
        return "NA"
    return ",".join(variant)



import sqlite3
from sqlite3 import Error


def create_connection():
    """ create a database connection to the SQLite database
        specified by the db_file
    :return: Connection object or None
    """
    try:
        conn = sqlite3.connect(":memory:")
        return conn
    except Error as e:
        print(e)
    return None


def create_reads_table(conn, key="sequence"):
    """ create a table on the SQLite database
    :param conn: connection of database
    :return: Connection object or None
    """
    c = conn.cursor()
    c.execute("CREATE TABLE reads"
              " (name text, sequence text,"
              " chrom text, start int,"
              " PRIMARY KEY(%s, chrom, start))" % key)
    conn.commit()


def insert_row_in_reads_table(cur, fields):
    """ create a table on the SQLite database
    :param cur: connection of database
    :param fields: list with columns to fill table
    :return: Connection object or None
    """
    # c = conn.cursor()
    cur.execute("INSERT INTO reads VALUES"
                " (\"%s\", \"%s\", \"%s\", %s)" % (fields[0],
                                                   fields[1],
                                                   fields[2],
                                                   fields[3]))


def select_all_reads(conn):
    """
    Query all rows in the reads table
    :param conn: the Connection object
    :return:
    """
    cur = conn.cursor()
    cur.execute("SELECT * FROM reads")

    rows = cur.fetchall()

    return rows



"""This directory is setup with configurations to run the main functional test.

Inspired in bcbio-nextgen code
"""
from __future__ import print_function
import os
import subprocess
import unittest
import shutil
import contextlib
import functools

from nose import SkipTest
from nose.plugins.attrib import attr


@contextlib.contextmanager
def make_workdir():
    remove_old_dir = True
    dirname = os.path.join(os.path.dirname(__file__), "test_automated_output")
    if remove_old_dir:
        if os.path.exists(dirname):
            shutil.rmtree(dirname)
        os.makedirs(dirname)
    orig_dir = os.getcwd()
    try:
        os.chdir(dirname)
        yield dirname
    finally:
        os.chdir(orig_dir)


def expected_failure(test):
    """Small decorator to mark tests as expected failure.
    Useful for tests that are work-in-progress.
    """
    @functools.wraps(test)
    def inner(*args, **kwargs):
        try:
            test(*args, **kwargs)
        except Exception:
            raise SkipTest
        else:
            raise AssertionError('Failure expected')
    return inner


class AutomatedAnalysisTest(unittest.TestCase):
    """Setup a full automated analysis and run the pipeline.
    """

    def setUp(self):
        self.data_dir = os.path.join(os.path.dirname(__file__), "data", "automated")

    def _install_test_files(self, data_dir):
        """Download required sequence and reference files.
        """
        #       self._download_to_dir(url, dirname)

    def _download_to_dir(self, url, dirname):
        print(dirname)
        cl = ["wget", url]
        subprocess.check_call(cl)
        cl = ["tar", "-xzvpf", os.path.basename(url)]
        subprocess.check_call(cl)
        shutil.move(os.path.basename(dirname), dirname)
        os.remove(os.path.basename(url))

    @attr(simulate=True)
    def test_simulate(self):
        """Check simulated data"""
        mirna = "TGAGGTAGTAGGTTGTATAGTT"
        correct = 0
        n = 0
        with open("data/examples/simulation/res/reads.mirna") as inh:
            header = inh.readline()
            for line in inh:
                cols = line.strip().split()
                mut, add, t5, t3 = cols[6:10]
                seq = cols[0]
                if mut!="0":
                    pos = int(mut[:-2])
                    nt1 = mut[-2]
                    nt2 = mut[-1]
                    seql = list(seq)
                    seql[pos] = nt2
                    seq = "".join(seql)
                if t5!="0" and t5.islower():
                    seq = "%s%s" % (t5.upper(), seq)
                elif t5!="0" and t5.isupper():
                    seq = seq[len(t5):]
                if add!="0":
                    seq = seq[:-len(add)]
                if t3!="0" and t3.islower():
                    seq = "%s%s" % (seq, t3.upper())
                elif t3!="0" and t3.isupper():
                    seq = seq[:-len(t3)]
                if seq == mirna:
                    correct += 1
                else:
                    print("\nerror:\n%s\n%s" % (seq, mirna))
                n += 1
        print("rate %s/%s" % (correct, n))

    @attr(complete=True)
    @attr(annotate=True)
    @attr(bam=True)
    @attr(cmd=True)
    def test_srnaseq_annotation_bam(self):
        """Run miraligner analysis
        """
        with make_workdir():
            clcode = ["mirtop",
                      "gff",
                      "--sps", "hsa", "--add-extra",
                      "--hairpin", "../../data/examples/annotate/hairpin.fa",
                      "--gtf", "../../data/examples/annotate/hsa.gff3",
                      "-o", "test_out_mirs",
                      "../../data/examples/annotate/sim_isomir.sam"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(low_memory=True)
    @attr(cmd=True)
    def test_srnaseq_annotation_bam_chunk(self):
        """Run miraligner analysis
        """
        with make_workdir():
            clcode = ["mirtop",
                      "gff", "--low-memory",
                      "--sps", "hsa", "--add-extra",
                      "--hairpin", "../../data/examples/annotate/hairpin.fa",
                      "--gtf", "../../data/examples/annotate/hsa.gff3",
                      "-o", "test_out_mirs",
                      "../../data/examples/annotate/sim_isomir.sam"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(cmd_bam_genomic=True)
    @attr(complete=True)
    @attr(cmd=True)
    def test_srnaseq_annotation_genomic_bam(self):
        """Run genomic bam analysis
        """
        with make_workdir():
            clcode = ["mirtop",
                      "gff",
                      "--sps", "hsa", "--add-extra", "--genomic",
                      "--hairpin", "../../data/examples/annotate/hairpin.fa",
                      "--gtf", "../../data/db/mirbase/hsa.gff3",
                      "-o", "test_out_mirs",
                      "../../data/examples/annotate/hsa-let-7a-nm.sam"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(cmd_bam_genomic_low_memory=True)
    @attr(complete=True)
    @attr(cmd=True)
    def test_srnaseq_annotation_genomic_bam_low_memory(self):
        """Run genomic bam analysis
        """
        with make_workdir():
            clcode = ["mirtop",
                      "gff", "--genomic", "--low-memory",
                      "--sps", "hsa", "--add-extra",
                      "--hairpin", "../../data/examples/annotate/hairpin.fa",
                      "--gtf", "../../data/db/mirbase/hsa.gff3",
                      "-o", "test_out_mirs",
                      "../../data/examples/annotate/hsa-let-7a-nm.sam"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(cmd_seqbuster=True)
    @attr(cmd=True)
    def test_srnaseq_annotation_seqbuster(self):
        """Run miraligner analysis
        """
        with make_workdir():
            clcode = ["mirtop",
                      "gff",
                      "--format", "seqbuster",
                      "--sps", "hsa",
                      "--hairpin", "../../data/examples/annotate/hairpin.fa",
                      "--gtf", "../../data/examples/annotate/hsa.gff3",
                      "-o", "test_out_mirs",
                      "../../data/examples/seqbuster/reads.mirna"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(cmd_seqbuster_low_memory=True)
    @attr(cmd=True)
    def test_srnaseq_annotation_seqbuster_low_memory(self):
        """Run miraligner analysis
        """
        with make_workdir():
            clcode = ["mirtop",
                      "gff", "--low-memory",
                      "--format", "seqbuster",
                      "--sps", "hsa",
                      "--hairpin", "../../data/examples/annotate/hairpin.fa",
                      "--gtf", "../../data/examples/annotate/hsa.gff3",
                      "-o", "test_out_mirs",
                      "../../data/examples/seqbuster/reads.mirna"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(cmd_isomirsea=True)
    @attr(cmd=True)
    def test_srnaseq_annotation_isomirsea(self):
        """Run isomirsea analysis
        """
        with make_workdir():
            clcode = ["mirtop",
                      "gff",
                      "--format", "isomirsea",
                      "--sps", "hsa",
                      "--hairpin", "../../data/examples/annotate/hairpin.fa",
                      "--gtf", "../../data/examples/annotate/hsa.gff3",
                      "-o", "test_out_mirs",
                      "../../data/examples/isomir-sea/tagMir-all.gff",
                      "-d", "-vd"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(cmd_srnabench=True)
    @attr(cmd=True)
    def test_srnaseq_annotation_srnabench(self):
        """Run srnabench analysis
        """
        with make_workdir():
            clcode = ["mirtop",
                      "gff",
                      "--format", "srnabench",
                      "--sps", "hsa",
                      "--hairpin", "../../data/examples/annotate/hairpin.fa",
                      "--gtf", "../../data/examples/annotate/hsa.gff3",
                      "-o", "test_out_mirs",
                      "../../data/examples/srnabench/",
                      "-d", "-vd"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(cmd_optimir=True)
    @attr(cmd=True)
    def test_srnaseq_annotation_optimir(self):
        """Run optimir analysis
        """
        with make_workdir():
            clcode = ["mirtop",
                      "gff",
                      "--format", "optimir",
                      "--sps", "hsa",
                      "--hairpin", "../../data/examples/annotate/hairpin.fa",
                      "--gtf", "../../data/examples/annotate/hsa.gff3",
                      "-o", "test_out_mirs",
                      "../../data/examples/optimir/synthetic_100_full.gff3",
                      "-d", "-vd"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(cmd_manatee=True)
    @attr(cmd=True)
    def test_srnaseq_annotation_manatee(self):
        """Run Manatee analysis
        """
        with make_workdir():
            clcode = ["mirtop",
                      "gff",
                      "--format", "manatee",
                      "--sps", "hsa",
                      "--hairpin", "../../data/examples/annotate/hairpin.fa",
                      "--gtf", "../../data/examples/annotate/hsa.gff3",
                      "-o", "test_out_mirs",
                      "../../data/examples/manatee/simulated.sam",
                      "-d", "-vd"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(cmd_stats=True)
    @attr(cmd=True)
    def test_srnaseq_stats(self):
        """Run stats analysis
        """
        with make_workdir():
            clcode = ["mirtop",
                      "stats",
                      "-o", "test_out_mirs",
                      "../../data/examples/gff/correct_file.gff"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(cmd_stats=True)
    @attr(cmd=True)
    def test_srnaseq_stats(self):
        """Run stats analysis
        """
        with make_workdir():
            clcode = ["mirtop",
                      "stats",
                      "../../data/examples/gff/correct_file.gff"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(cmd_merge=True)
    @attr(cmd=True)
    def test_merge_bam(self):
        """
        Run collapse two samples
        """
        with make_workdir():
            clcode = ["mirtop",
                      "gff",
                      "--sps", "hsa", "--add-extra",
                      "--hairpin", "../../data/examples/annotate/hairpin.fa",
                      "--gtf", "../../data/examples/annotate/hsa.gff3",
                      "-o", "test_out_mirs",
                      "../../data/merge/samples1.sam",
                      "../../data/merge/samples2.sam"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(cmd_export_seqbuster=True)
    @attr(cmd=True)
    def test_export_seqbuster(self):
        """
        Run SEQBUSTER export command
        """
        with make_workdir():
            clcode = ["mirtop",
                      "export",
                      "-o", "test_out_mirs",
                      "--hairpin", "../../data/examples/annotate/hairpin.fa",
                      "--gtf", "../../data/examples/annotate/hsa.gff3",
                      "../../data/examples/gff/correct_file.gff"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(cmd_export_vcf=True)
    @attr(cmd=True)
    def test_export_vcf(self):
        """
        Run VCF export command
        """
        with make_workdir():
            clcode = ["mirtop",
                      "export",
                      "-o", "test_out_mirs",
                      "--format", "vcf",
                      "-d", "-vd",
                      "--hairpin", "../../data/examples/annotate/hairpin.fa",
                      "--gtf", "../../data/examples/annotate/hsa.gff3",
                      "../../data/examples/gff/correct_file.gff"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(cmd_export_fasta=True)
    @attr(cmd=True)
    def test_export_fasta(self):
        """
        Run FASTA export command
        """
        with make_workdir():
            clcode = ["mirtop",
                      "export",
                      "-o", "test_out_mirs",
                      "--format", "fasta",
                      "-d", "-vd",
                      "--hairpin", "../../data/examples/annotate/hairpin.fa",
                      "--gtf", "../../data/examples/annotate/hsa.gff3",
                      "../../data/examples/gff/correct_file.gff"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(cmd_count=True)
    @attr(cmd=True)
    def test_count(self):
        """
        Run count command
        """
        with make_workdir():
            clcode = ["mirtop",
                      "counts",
                      "-o", "test_out_mirs",
                      "--hairpin", "../../data/examples/annotate/hairpin.fa",
                      "--gtf", "../../data/examples/annotate/hsa.gff3",
                      "--gff", "../../data/examples/synthetic/let7a-5p.gff"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(cmd_spikeins=True)
    @attr(cmd=True)
    def test_spikeins_cmd(self):
        """Run spikeins analysis
        """
        import platform
        with make_workdir():
            shutil.copy("../../data/examples/spikeins/spikeins.fa",
                        "spikeins.fa")
            clcode = ["mirtop",
                      "spikein",
                      "spikeins.fa",
                      "-o",
                      "test_out_spikeins"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

            if platform.system() == "Linux":
                clcode = ["razers3", "-dr", "0", "-i", "80", "-rr", "90",
                          "-f", "-o", "spikeins.sam",
                          "test_out_spikeins/spikeins_pre.fasta",
                          "../../data/examples/spikeins/test-spikeins.fa"]
                print(" ".join(clcode))
                subprocess.check_call(clcode)
            else:
                shutil.copy("../../data/examples/spikeins/spikeins.sam",
                            "spikeins.sam")
            clcode = ["mirtop",
                      "gff",
                      "--add-extra",
                      "--hairpin", "test_out_spikeins/spikeins_pre.fasta",
                      "--gtf", "test_out_spikeins/spikeins_pre.gff",
                      "-o", "test_out_mirs",
                      "spikeins.sam"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(cmd_update=True)
    @attr(cmd=True)
    def test_update_cmd(self):
        """Run update analysis
        """
        with make_workdir():
            clcode = ["mirtop",
                      "update",
                      "-o", "test_out_mirs",
                      "../../data/examples/versions/version1.0.gff"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)

    @attr(complete=True)
    @attr(cmd_validate=True)
    @attr(cmd=True)
    def test_validate_cmd(self):
        """Run update analysis
        """
        with make_workdir():
            clcode = ["mirtop",
                      "validate",
                      "../../data/examples/gff/correct_file.gff"]
            print("")
            print(" ".join(clcode))
            subprocess.check_call(clcode)